# -*- coding: utf-8 -*-
"""
Conversion between Russian National Corpus (www.ruscorpora.com)
tagset and other tagsets (currently OpenCorpora tagset).

Limitations:

* OpenCorpora has grammemes for active/passive voices; RNC
  has active/passive/middle voices. "Passive" is reliable,
  "active/middle" are not.

* tagsets differ in pronoun/adverb/numeral/predicative/transitions handling;
  this module tries not to be too smart: it only applies unambiguous
  transformations;

* Abbr mean different things in OpenCorpora and RNC;

* RNC's adnum is not the same as OpenCorpora's Coun;

* OpenCorpora Comparative is always converted to A, while it may be
  A, ADV, NUM and PRAEDIC in RNC (there is not enough information to make
  the 1-to-1 conversion);

"""

from __future__ import absolute_import, unicode_literals
from russian_tagsets import rule_engine
from russian_tagsets import converters

FROM_OPENCORPORA = rule_engine.parse("""
# part of speech
NOUN => S

ADJF,Apro,Subx,Anph => A-PRO
ADJF,Apro,Subx => S-PRO
ADJF,Apro => A-PRO
ADJF,Anum => ANUM
ADJF => A,plen
ADJS => A,brev

COMP,Cmp2 => A,comp2
COMP,Supr => A,supr
COMP => A,comp

# INFN => V,inf,med
# INFN => V,inf,act
INFN => V,inf

PRTF,inan => V,partcp,plen
PRTF,anim => V,partcp,plen
PRTF => V,partcp,plen

PRTS,intr => V,intr,partcp,plen
PRTS => V,tran,partcp,brev

GRND => V,ger

VERB,impr,incl,sing => V,1p,imper,pl
VERB,impr,incl,plur => V,1p,imper2,pl
VERB,impr,excl => V,2p,imper
VERB,impr,incl => V,1p,imper

VERB => V

NUMR,inan => NUM
NUMR => NUM

PRED,pres => PRAEDIC
PRED => PRAEDIC

ADVB,Ques => ADV-PRO
ADVB => ADV

NPRO => S-PRO
PREP => PR
PRCL => PART
INTJ => INTJ

CONJ,Prnt => PARENTH
CONJ => CONJ

# animacy
anim => anim
inan => inan

# gender
femn,Ms-f => m-f
masc,Ms-f => m-f
neut,Ms-f => m-f
Ms-f => m-f

masc => m
femn => f
neut => n

# number
sing => sg
plur => pl
Fixd => 0

# case
nomn => nom
gent => gen
datv => dat
accs => acc
ablt => ins
loct => loc
voct => voc
gen1 => gen
gen2 => gen2
acc2 => acc2
loc1 => loc
loc2 => loc2

# Abbr doesn't mean the same
# Abbr => abbr

Name => persn
Surn => famn
Patr => patrn
Dist => distort
Arch => anom

# ??
Litr => anom
# Infr => anom
# V-oy => anom
V-be => distort

Supr => supr

# --
# ??
Coun => adnum

# aspect
perf => pf
impf => ipf

# transitivity
tran => tran
intr => intr

# voice
actv => act
pssv => pass

# person
1per => 1p
2per => 2p
3per => 3p

# tense
pres => praes
past => praet
futr => fut

# mood
indc => indic
impr => imper

# extra grammemes
LATN => NONLEX
ROMN => ANUM,ciph

# non-standard & pymorphy2-specific
# (RNC doesn't tag punctuation marks)
PNCT => PNCT
NUMB => NUM,ciph

# hack to preserve whitespace info:
| => =
""")


def _is_initials(opencorpora_grammemes):
    if set(['NOUN', 'Sgtm', 'Abbr', 'Fixd']).issubset(opencorpora_grammemes):
        if set(['Name', 'Patr']).intersection(opencorpora_grammemes):
            return True
    return False


def from_opencorpora_int(open_tag):
    """
    Convert OpenCorpora tag to www.ruscorpora.com tag::

        >>> print(from_opencorpora_int('NOUN,inan,masc sing,nomn'))
        S,inan,m=sg,nom

    """

    # Whitespace is replaced with ",|,"
    # then "|" is treated as token and replaced with "=",
    # then commas around "=" are removed in result.
    # This way space is converted to "=".

    grammeme_list = open_tag.replace(' ', ',|,').split(',')

    if _is_initials(grammeme_list):
        return 'INIT=abbr'

    result = rule_engine.apply_rules(FROM_OPENCORPORA, grammeme_list)
    result = ','.join(result).replace(',=,', '=').replace(',=', '')
    if result == '':
        return 'NONLEX'
    return result


converters.add('opencorpora-int', 'ruscorpora', from_opencorpora_int)



# -*- coding: utf-8 -*-
"""
Conversion between aot.ru/pymorphy and positional
(http://ufal.mff.cuni.cz/~hana/morph/rutags.html) tags.

Both "aot -> positional" and "positional -> aot" conversions discard
some information because of tagset differences.

AOT tags are generally less detailed than positional tags so
"aot -> positional" conversion is about safe and
"positional -> aot" conversion is more lossy.

"aot -> positional" issues
--------------------------

* "безл", "указат", "вопр", "мр-жр", "имя", "отч" tags are discarded;
* "МС-ПРЕДК", "ПРЕДК" tags are lost in "aot -> positional -> aot"
  conversion (they become "Н");
* animacy is sometimes discarded.

"positional -> aot issues"
--------------------------

* possessor's gender, possessor's number, reflexivity, verbal aspect and
  negation are discarded because they don't have counterparts in aot tagset;
* pronounce, numeral, conjunction, preposition and
  participle classification is much simpler in aot;
* punctuation is not supported;
* possession information for adjectives is discarded;
* adverb classification criteria is different in aot;
* 'X' ("any number") number is not supported.

"""

from __future__ import absolute_import, unicode_literals
from russian_tagsets import converters
from russian_tagsets import positional
from russian_tagsets.utils import invert_mapping

def split_tag_raw(aot_tag):
    if ',' in aot_tag:
        return aot_tag.split(',', 1)
    else:
        return aot_tag, ''

def split_tag(aot_tag):
    pos, info = split_tag_raw(aot_tag)
    if info:
        return pos, set(info.split(','))
    else:
        return pos, set()

def join_tag(pos, info_set):
    return  ",".join([pos] + list(info_set))

_POS_MAP = {
    'N': 'С',

    'A': 'П',
    'AG': 'ПРИЧАСТИЕ',
    'Ac': 'КР_ПРИЧАСТИЕ',
    'AC': 'КР_ПРИЛ',

    'P': 'МС',
    'Pq': 'МС-П',
    'PS': 'МС-П',
    'Pw': 'МС-П',
    'Pz': 'МС-П',

    'C': 'ЧИСЛ',
    'Cr': 'ЧИСЛ-П',

    'V': 'Г',
    'VB': 'Г',
    'Vi': 'Г,пвл',
    'Ve': 'ДЕЕПРИЧАСТИЕ',
    'Vf': 'ИНФИНИТИВ',

    'D': 'Н',

    'R': 'ПРЕДЛ',
    'RF': 'ФРАЗ',
    'J': 'СОЮЗ',
    'I': 'МЕЖД',
    'T': 'ЧАСТ',
    'Z': 'ЗНАК', # non-standard
}

TENSES = {
    'нст': 'P',
    'прш': 'R',
    'буд': 'F',
}
TENSES_INV = invert_mapping(TENSES)

CASES = {
    'им': '1',
    'рд': '2',
    'дт': '3',
    'вн': '4',
    'тв': '7',
    'пр': '6',
}
CASES_INV = invert_mapping(CASES)

GENDERS = {
    'мр': 'M',
    'жр': 'F',
    'ср': 'N',
}
GENDERS_INV = invert_mapping(GENDERS)

PERSONS = {'1л': '1', '2л': '2', '3л': '3'}
PERSONS_INV = invert_mapping(PERSONS)


def to_positional(aot_tag):
    """
    Convert AOT.ru tag to positional.Tag format.
    This is lossy because of format differences.
    """

    pos, info = split_tag(aot_tag)

    tag = positional.Tag()

    # ==== 1,2: POS tag ====
    if pos == 'С':
        tag.POS = 'NN'

    elif pos == 'П':
        tag.POS = 'AA'
        # lossy: possessive adjective (мамин, овечью) should be AU
        # some comparative adjectives should be adverbs

    elif pos == 'Г':
        tag.POS = 'VB'
        if 'пвл' in info:
            tag.POS = 'Vi'

    elif pos == 'ПРИЧАСТИЕ':
        # Participle, active or long passive
        # (читающий, читавший, прочитавший, читаемый;
        # but not прочитанный (AA), прочитан (Ac)
        tag.POS = 'AG'

    elif pos == 'КР_ПРИЧАСТИЕ':
        tag.POS = 'Ac'

    elif pos == 'КР_ПРИЛ':
        tag.POS = 'AC'

    elif pos == 'ДЕЕПРИЧАСТИЕ':
        tag.POS = 'Ve'

    elif pos == 'ИНФИНИТИВ':
        tag.POS = 'Vf'

    elif pos == 'МС':
        # Personal pronoun (ja, my, ty, vy, on, ona, ono, oni, sebja)
        tag.POS = 'PP'

        # fixme
        # 3rd person pronoun in prepositional forms (nego, nej, ...)
        # tag.POS = 'P5'

        # Relative/interrogative pronoun with nominal declension (kto, čto)
        # tag.POS = 'PQ'

        # Negative pronoun with nominal declension (nicto, nikto)
        # tag.POS = 'PW'

        # Indefinite pronoun with nominal declension (kto-to, kto-nibud', cto-to, ...)
        # tag.POS = 'PZ'

        # Pronoun demonstrative (ètot, tot, sej, takoj, èkij, ... )
        # tag.POS = 'PD'

    elif pos == 'МС-ПРЕДК':
        # e.g. нечего
        tag.POS = 'Db' # lossy

    elif pos == 'МС-П':
        # e.g. всякий

        # Relative/interrogative pronoun with adjectival
        # declension (kakoj, kotoryj, cej, ...)
        tag.POS = 'Pq'

        # fixme/hack: Possessive pronoun (moj, ego, svoj, ..)
        if 'од' in info:
            tag.POS = 'PS'

        # fixme: Negative pronoun with adjectival declension (nikakoj, nicej)
        # tag.POS = 'Pw'

        # fixme: Indefinite pronoun with adjectival declension (samyj, ves', ...)
        # tag.POS = 'Pz'

    elif pos == 'Н':
        # Adverb without a possibility to form negation
        # and degrees of comparison (vverxu, vnizu, potom)
        tag.POS = 'Db'

#        if 'указат' in info or 'вопр' in info:
#            tag.POS = 'Db'
#        else:
#            tag.POS = 'Dg'

        # fixme: adverb forming negation and degrees of
        # comparison (vysoko, daleko). Idea: pymorphy.get_graminfo
        # returns 'КР_ПРИЛ' and 'Н' variants for such adverbs.
        # tag.POS == 'Dg'

        # fixme: Multiplicative numeral (dvaždy, triždy) are adverbs

    elif pos == 'ПРЕДК':
        # e.g. интересно
        tag.POS = 'Dg' # ?

    elif pos == 'ЧИСЛ':
        # Cardinal numeral (odin, tri, sorok)
        tag.POS = 'Cn'

        # fixme: Generic/collective numeral (dvoje, četvero)
        # tag.POS = 'Cj'

        # Indefinite numeral (mnogo, neskol'ko)
        # tag.POS = 'Ca'

        # Interrogative numeral (skol'ko)
        # tag.POS = 'Cu'

        # Multiplicative numeral (dvaždy, triždy)
        # tag.POS = 'Cv'

    elif pos == 'ЧИСЛ-П':
        tag.POS = 'Cr'

    elif pos == 'ПРЕДЛ':
        # Nonvocalized preposition (ob, pered, s, v, ...)
        tag.POS = 'RR'

        # fixme: Vocalized preposition (obo, peredo, so, vo, ...)
        # tag.POS = 'RV'

    elif pos == 'СОЮЗ':
        # Non-subordinate conjunction (i, a, xotja, pricem)
        tag.POS = 'J^'
        # fixme: Subordinate conjunction (esli, čto, kotoryj)
        # tag.POS = 'J,'

    elif pos == 'МЕЖД':
        tag.POS = 'II'

    elif pos == 'ЧАСТ':
        tag.POS = 'TT'

    elif pos == 'ВВОДН':
        # e.g. конечно
        tag.POS = 'Db' # ?

    elif pos == 'ФРАЗ':
        # e.g. несмотря
        tag.POS = 'RF'

    # ==== 3: gender ====

    for gender in GENDERS:
        if gender in info:
            tag.gender = GENDERS[gender]
            break
    else:
        if 'мр-жр' in info:
            tag.gender = 'X' # fixme?

    # ==== 4: animacy ====
    # FIXME
    # X should be used: except for nouns, in all
    # forms except accusative masculine singular and accusative
    # plural of all genders. Non-declinable words in all forms.
    if info.issuperset(['од', 'но']):
        tag.animacy = 'X' # fixme
    elif 'од' in info:
        tag.animacy = 'A'
    elif 'но' in info:
        tag.animacy = 'I'
    else:
        tag.animacy = '-' # fixme?

    # ======== 5: number =======
    if 'ед' in info:
        tag.number = 'S'
    elif 'мн' in info:
        tag.number = 'P'
    # FIXME
    # non-declinable nouns, adjectives and verbs,
    # 3rd person possessive pronouns:
    # tag.number = 'X'

    # ======== 6: case ==========
    for case in CASES:
        if case in info:
            tag.case = CASES[case]
        if '2' in info:
            tag.variant = '1'
    if 'зв' in info:
        tag.case = '1'
        tag.variant = '1'

    if '0' in info:
        tag.case = 'X'
        tag.number = 'X'

    # TODO: 7: possessor's gender
    # TODO: 8: possessor's number

    # ======= 9: person ========
    # fixme: it should be X for non-declinable verbs
    for person in PERSONS:
        if person in info:
            tag.person = PERSONS[person]
            break

    # ======= 10: reflexivity ========
    if tag.POS in ['AG', 'PP', 'P5', 'PS'] or tag.mainPOS == 'V':
        tag.reflexivity = 'I'
        # fixme: lossy! e.g. for "получиться" it should be 'R'.

    # ========= 15: voice ===========
    # XXX: this is out of order because tense needs it
    if tag.POS in ['AG', 'Ac']: # ??
        if 'дст' in info:
            tag.voice = 'A'
        elif 'стр' in info:
            tag.voice = 'P'

    # ======= 11: tense =========
    if tag.POS in ['AG', 'VB', 'Vp']: # ?
        for tense in TENSES:
            if tense in info:
                tag.tense = TENSES[tense]
                break
        # passive long participles
        if tag.POS == 'AG' and tag.tense == '-' and tag.voice == 'P':
            tag.tense = 'X'


    # TODO: 12: verbal_aspect

    # ========== 13: degree of comparison =========
    if tag.POS in ['AA', 'Dg']:
        if 'сравн' in info:
            tag.degree_of_comparison = '2'
        elif 'прев' in info:
            tag.degree_of_comparison = '3'
        else:
            tag.degree_of_comparison = '1'

    # ========= 14: negation =========
    if tag.mainPOS in ['N', 'A'] or tag.POS == 'Dg':
        tag.negation = 'A'
        # fixme: lossy!

    # ======== 16: variant ===========
    if 'арх' in info:
        tag.variant = '2'
    elif 'разг' in info:
        tag.variant = '5'
    elif 'аббр' in info:
        tag.variant = '8'

    return tag


def from_positional(positional_tag):
    """
    Convert positional.Tag to AOT format.
    This is lossy because of format differences.
    """
    if not isinstance(positional_tag, positional.Tag):
        positional_tag = positional.Tag(positional_tag)

    tag = positional_tag
    pos, info = '', set()

    if tag.POS in _POS_MAP:
        pos = _POS_MAP[tag.POS]
    elif tag.mainPOS in _POS_MAP:
        pos = _POS_MAP[tag.mainPOS]
    else:
        pos = tag.mainPOS

    # ==== 3. gender ======
    if tag.gender in GENDERS_INV:
        info.add(GENDERS_INV[tag.gender])

    # ==== 4. animacy =====
    if tag.animacy == 'A':
        info.add('од')
    elif tag.animacy == 'I':
        info.add('но')
    elif tag.animacy == 'X':
        info.update(['од', 'но'])

    # ===== 5. number =======
    if tag.number == 'P':
        info.add('мн')
    elif tag.number == 'S':
        info.add('ед')
    elif tag.number == 'X':
        pass # fixme

    # ====== 6. case =========
    if tag.case in CASES_INV:
        if tag.variant == '1':
            if tag.case == '1':
                info.add('зв')
            else:
                info.add(CASES_INV[tag.case])
                info.add('2')
        else:
            info.add(CASES_INV[tag.case])

    if tag.case == 'X' and tag.number == 'X':
        info.add('0')

    # 7. possessor's gender
    # 8. possessor's number

    # ====== 9. person ========
    if tag.person in PERSONS_INV:
        info.add(PERSONS_INV[tag.person])

    # 10. reflexivity

    # ====== 11. tense ========
    if tag.tense in TENSES_INV:
        info.add(TENSES_INV[tag.tense])

    # 12. verbal aspect

    # ====== 13. degree of comparison ======
    if tag.degree_of_comparison == '2':
        info.add('сравн')
    elif tag.degree_of_comparison == '3':
        info.add('прев')
    if tag.POS == 'Dg' and tag.degree_of_comparison == '2': # hack?
        pos = 'П'

    # 14. negation

    # ====== 15. voice ==========
    if tag.voice == 'A':
        info.add('дст')
    elif tag.voice == 'P':
        info.add('стр')
    elif tag.mainPOS == 'V' and tag.tense != 'F': # hack?
        info.add('дст')

    # ====== 16. variant ========
    if tag.variant in ['2', '3']:
        info.add('арх')
    elif tag.variant in ['5', '6', '7']:
        info.add('разг')
    elif tag.variant == '8':
        info.add('аббр')

    return ",".join([pos] + list(info))

converters.add('positional', 'aot', from_positional)
converters.add('aot', 'positional', to_positional)


# -*- coding: utf-8 -*-
from __future__ import absolute_import

def invert_mapping(dct):
    return dict((v, k) for k,v in dct.items())

def bigrams(iterable):
    _first = object()
    prev = _first
    for current in iterable:
        if prev is not _first:
            yield prev, current
        prev = current

def find_shortest_path(graph, start, end, cost=lambda path: len(path), path=[]):
    # adopted from http://www.python.org/doc/essays/graphs/
    path = path + [start]
    if start == end:
        return path
    if start not in graph:
        return None
    shortest = None
    for node in graph[start]:
        if node not in path:
            newpath = find_shortest_path(graph, node, end, cost, path)
            if newpath:
                if not shortest or cost(newpath) < cost(shortest):
                    shortest = newpath
    return shortest




# -*- coding: utf-8 -*-
"""
Conversion between OpenCorpora.org (http://opencorpora.org/dict.php?act=gram)
tagset and other tagsets.
"""
from __future__ import absolute_import, unicode_literals
from russian_tagsets import converters, positional, aot
from russian_tagsets.utils import invert_mapping

GRAM_TABLE = [
    # num, internal tag, external tag, description, parent, aot.ru tag
    ['1', 'POST', 'ЧР', 'часть речи', '—', None],
    ['2', 'NOUN', 'СУЩ', 'имя существительное', 'POST', 'С'],
    ['3', 'ADJF', 'ПРИЛ', 'имя прилагательное (полное)', 'POST', 'П'],
    ['4', 'ADJS', 'КР_ПРИЛ', 'имя прилагательное (краткое)', 'POST', 'КР_ПРИЛ'],
    ['5', 'COMP', 'КОМП', 'компаратив', 'POST', 'П'], # fixme: aot
    ['6', 'VERB', 'ГЛ', 'глагол (личная форма)', 'POST', 'Г'],
    ['7', 'INFN', 'ИНФ', 'глагол (инфинитив)', 'POST', 'ИНФИНИТИВ'],
    ['8', 'PRTF', 'ПРИЧ', 'причастие (полное)', 'POST', 'ПРИЧАСТИЕ'],
    ['10', 'PRTS', 'КР_ПРИЧ', 'причастие (краткое)', 'POST', 'КР_ПРИЧАСТИЕ'],
    ['11', 'GRND', 'ДЕЕПР', 'деепричастие', 'POST', 'ДЕЕПРИЧАСТИЕ'],
    ['12', 'NUMR', 'ЧИСЛ', 'числительное', 'POST', 'ЧИСЛ'],
    ['13', 'ADVB', 'Н', 'наречие', 'POST', 'Н'],
    ['14', 'NPRO', 'МС', 'местоимение-существительное', 'POST', 'МС'],
    ['15', 'PRED', 'ПРЕДК', 'предикатив', 'POST', 'ПРЕДК'],
    ['16', 'PREP', 'ПР', 'предлог', 'POST', 'ПРЕДЛ'],
    ['17', 'CONJ', 'СОЮЗ', 'союз', 'POST', 'СОЮЗ'],
    ['18', 'PRCL', 'ЧАСТ', 'частица', 'POST', 'ЧАСТ'],
    ['19', 'INTJ', 'МЕЖД', 'междометие', 'POST', 'МЕЖД'],

    ['21', 'ANim', 'Од-неод', 'одушевлённость / одушевлённость не выражена', '—', None],
    ['22', 'anim', 'од', 'одушевлённое', 'ANim', 'од'],
    ['23', 'inan', 'неод', 'неодушевлённое', 'ANim', 'но'],

    ['24', 'GNdr', 'хр', 'род / род не выражен', '—', None],
    ['25', 'masc', 'мр', 'мужской род', 'GNdr', 'мр'],
    ['26', 'femn', 'жр', 'женский род', 'GNdr', 'жр'],
    ['27', 'neut', 'ср', 'средний род', 'GNdr', 'ср'],
    ['28', 'Ms-f', 'ор', 'общий род', '—', 'мр-жр'],

    ['29', 'NMbr', 'Число', 'число', '—', None],
    ['30', 'sing', 'ед', 'единственное число', 'NMbr', 'ед'],
    ['31', 'plur', 'мн', 'множественное число', 'NMbr', 'мн'],
    ['32', 'Sgtm', 'sg', 'singularia tantum', '—', None], # fixme: aot?
    ['33', 'Pltm', 'pl', 'pluralia tantum', '—', None], # fixme: aot?
    ['36', 'Fixd', '0', 'неизменяемое', '—', '0'],

    ['37', 'CAse', 'Падеж', 'категория падежа', '—', None],
    ['38', 'nomn', 'им', 'именительный падеж', 'CAse', 'им'],
    ['39', 'gent', 'рд', 'родительный падеж', 'CAse', 'рд'],
    ['40', 'datv', 'дт', 'дательный падеж', 'CAse', 'дт'],
    ['41', 'accs', 'вн', 'винительный падеж', 'CAse', 'вн'],
    ['42', 'ablt', 'тв', 'творительный падеж', 'CAse', 'тв'],
    ['43', 'loct', 'пр', 'предложный падеж', 'CAse', 'пр'],
    ['44', 'voct', 'зв', 'звательный падеж', 'nomn', 'зв'],
    ['45', 'gen1', 'рд1', 'первый родительный падеж', 'gent', 'рд'], # fixme: aot
    ['46', 'gen2', 'рд2', 'второй родительный (частичный) падеж', 'gent', 'рд'], # fixme: aot
    ['47', 'acc2', 'вн2', 'второй винительный падеж', 'accs', 'вн'], # fixme: aot
    ['48', 'loc1', 'пр1', 'первый предложный падеж', 'loct', 'пр'], # fixme: aot
    ['49', 'loc2', 'пр2', 'второй предложный (местный) падеж', 'loct', 'пр'], # fixme: aot

    ['50', 'Abbr', 'аббр', 'аббревиатура', '—', 'аббр'],
    ['51', 'Name', 'имя', 'имя', '—', 'имя'],
    ['52', 'Surn', 'фам', 'фамилия', '—', 'фам'],
    ['53', 'Patr', 'отч', 'отчество', '—', 'отч'],
    ['54', 'Geox', 'гео', 'топоним', '—', 'лок'], # XXX: is aot tag correct?
    ['55', 'Orgn', 'орг', 'организация', '—', None],
    ['56', 'Trad', 'tm', 'торговая марка', '—', None],
    ['57', 'Subx', 'субст?', 'возможна субстантивация', '—', None],
    ['58', 'Supr', 'превосх', 'превосходная степень', '—', 'прев'],
    ['59', 'Qual', 'кач', 'качественное', '—', 'кач'],
    ['60', 'Apro', 'мест-п', 'местоименное', '—', None],
    ['61', 'Anum', 'числ-п', 'порядковое', '—', None],
    ['62', 'Poss', 'притяж', 'притяжательное', '—', 'притяж'],
    ['63', 'V-ey', '*ею', 'форма на -ею', '—', None],
    ['64', 'V-oy', '*ою', 'форма на -ою', '—', None],
    ['65', 'Cmp2', 'сравн2', 'сравнительная степень на по-', '—', None], # fixme: aot
    ['66', 'V-ej', '*ей', 'форма компаратива на -ей', '—', None], # fixme: aot

    ['67', 'ASpc', 'Вид', 'категория вида', '—', None],
    ['68', 'perf', 'сов', 'совершенный вид', 'ASpc', None],
    ['69', 'impf', 'несов', 'несовершенный вид', 'ASpc', None],

    ['70', 'TRns', 'Перех', 'категория переходности', '—', None],
    ['71', 'tran', 'перех', 'переходный', 'TRns', None],
    ['72', 'intr', 'неперех', 'непереходный', 'TRns', None],

    ['73', 'Impe', 'безл', 'безличный', '—', 'безл'],
    ['74', 'Uimp', 'безл-у', 'безличное употребление', '—', 'безл'], # fixme: aot?
    ['75', 'Mult', 'мног', 'многократный', '—', None],
    ['76', 'Refl', 'возвр', 'возвратный', '—', None],

    ['77', 'PErs', 'Лицо', 'категория лица', '—', None],
    ['78', '1per', '1л', '1 лицо', 'PErs', '1л'],
    ['79', '2per', '2л', '2 лицо', 'PErs', '2л'],
    ['80', '3per', '3л', '3 лицо', 'PErs', '3л'],

    ['81', 'TEns', 'Время', 'категория времени', '—', None],
    ['82', 'pres', 'наст', 'настоящее время', 'TEns', 'нст'],
    ['83', 'past', 'прош', 'прошедшее время', 'TEns', 'прш'],
    ['84', 'futr', 'буд', 'будущее время', 'TEns', 'буд'],

    ['85', 'MOod', 'Накл', 'категория наклонения', '—', None],
    ['86', 'indc', 'изъяв', 'изъявительное наклонение', 'MOod', None],
    ['87', 'impr', 'повел', 'повелительное наклонение', 'MOod', 'пвл'],

    ['88', 'INvl', 'Совм', 'категория совместности', '—', None],
    ['89', 'incl', 'вкл', 'говорящий включён в действие', 'INvl', None],
    ['90', 'excl', 'выкл', 'говорящий не включён в действие', 'INvl', None],

    ['91', 'VOic', 'Залог', 'категория залога', '—', None],
    ['92', 'actv', 'действ', 'действительный залог', 'VOic', 'дст'],
    ['93', 'pssv', 'страд', 'страдательный залог', 'VOic', 'стр'],

    ['94', 'Infr', 'разг', 'разговорное', '—', 'разг'],
    ['95', 'Slng', 'жарг', 'жаргонное', '—', 'жарг'],
    ['96', 'Arch', 'арх', 'устаревшее', '—', 'арх'],
    ['97', 'Litr', 'лит', 'литературный вариант', '—', None],
    ['98', 'Erro', 'опеч', 'опечатка', '—', None],
    ['99', 'Dist', 'искаж', 'искажение', '—', None],
    ['100', 'Ques', 'вопр', 'вопросительное', '—', 'вопр'],
    ['101', 'Dmns', 'указ', 'указательное', '—', 'указат'],
    ['103', 'Prnt', 'вводн', 'вводное слово', '—', None], # fixme: ВВОДН
    ['104', 'V-be', '*ье', 'форма на -ье', '—', None],
    ['105', 'V-en', '*енен', 'форма на -енен', '—', None],
    ['106', 'V-ie', '*ие', 'отчество через -ие-', '—', None],
    ['107', 'V-bi', '*ьи', 'форма на -ьи', '—', None],
    ['108', 'Fimp', '*несов', 'деепричастие от глагола несовершенного вида', '—', None], # fixme: aot
    ['109', 'Prdx', 'предк?', 'может выступать в роли предикатива', '—', None], # fixme: aot
    ['110', 'Coun', 'счетн', 'счётная форма', '—', None],
    ['111', 'Coll', 'собир', 'собирательное числительное', '—', None],
    ['112', 'V-sh', '*ши', 'деепричастие на -ши', '—', None],
    ['113', 'Af-p', '*предл', 'форма после предлога', '—', None],
    ['114', 'Inmx', 'неодуш?', 'может использоваться как неодушевлённое', '—', None],
    ['115', 'Vpre', 'в_предл', 'Вариант предлога ( со, подо, ...)',	'—', None],
    ['116', 'Anph', 'Анаф', 'Анафорическое (местоимение)', '-', None],
]

INTERNAL_TO_EXTERNAL = dict((item[1], item[2]) for item in GRAM_TABLE)
EXTERNAL_TO_INTERNAL = invert_mapping(INTERNAL_TO_EXTERNAL)

EXTERNAL_TO_AOT = dict((item[2], item[5]) for item in GRAM_TABLE
                        if item[5] is not None)

def _translate_comma_separated(tag_part, mapping):
    return ",".join([mapping.get(tok, tok).strip() for tok in tag_part.split(',')])

def _translate_tag(tag, mapping):
    return " ".join([
        _translate_comma_separated(part, mapping)
        for part in tag.split()
    ])


def external_to_internal(external_tag):
    return _translate_tag(external_tag, EXTERNAL_TO_INTERNAL)

def internal_to_external(internal_tag):
    return _translate_tag(internal_tag, INTERNAL_TO_EXTERNAL)


def to_aot(open_tag):
    open_tags = open_tag.replace(" ", ',').split(',')
    open_pos, open_info = open_tags[0], open_tags[1:]

    pos = EXTERNAL_TO_AOT.get(open_pos, None)
    info = [EXTERNAL_TO_AOT.get(tag, None) for tag in open_info]
    info = [tag for tag in info if tag is not None]

    if open_pos == 'ПРИЛ':
        if 'мест-п' in open_info:
            pos = 'МС-П'
        elif 'числ-п' in open_info:
            pos = 'ЧИСЛ-П'

    if 'пвл' in info:
        info.append('2л')

    if not pos:
        return ''
    return ','.join([pos] + info)


converters.add('opencorpora-int', 'opencorpora-ext', internal_to_external)
converters.add('opencorpora-ext', 'opencorpora-int', external_to_internal)
converters.add('opencorpora-ext', 'aot', to_aot)
